from celescope.tools.multi import Multi, TOOLS_DIR
from celescope.flv_CR.__init__ import __ASSAY__
from celescope.flv_CR.assemble import __SUB_STEPS__


class Multi_flv_CR(Multi):
    """

    ## Download and unpack cellranger soft and reference file.
    ```
    wget -O cellranger-7.0.1.tar.gz "https://cf.10xgenomics.com/releases/cell-vdj/cellranger-7.0.1.tar.gz?Expires=1668447380&Policy=eyJTdGF0ZW1lbnQiOlt7IlJlc291cmNlIjoiaHR0cHM6Ly9jZi4xMHhnZW5vbWljcy5jb20vcmVsZWFzZXMvY2VsbC12ZGovY2VsbHJhbmdlci03LjAuMS50YXIuZ3oiLCJDb25kaXRpb24iOnsiRGF0ZUxlc3NUaGFuIjp7IkFXUzpFcG9jaFRpbWUiOjE2Njg0NDczODB9fX1dfQ__&Signature=bZoM-0W1ExsZWYVhUJ80oSzKEaYTdlf1NoFiKLswTLWGveefjH1fJOkd1c4kMxCiTDfohyDpNWzk-xBMBme-u3r-0X7WFYSvfCdMyo2CSM4K~Ur73Sn30REr0kr7oC9byrlLMqG2mtO7DLfprDrAZGZDkfyDLpGJ8hb1qWSsWqRo8CxbqRzA69h0v65Qn86HMrHAeotdhxUVmBIvONwPmsC90J9K4gVDw1sDF39F4f89zguTGgSAY6mUdPG1cvHHZMeuLaJZDqRgODysOhsB-keYXW8cYa-R5chh9s1ASC4yA1QKSe-fBdB1-FLQdAwMjNEHdd7uGCWzLj4J~AXgJg__&Key-Pair-Id=APKAI7S6A5RYOXBWRPDA"

    Reference: human and mouse
    wget https://cf.10xgenomics.com/supp/cell-vdj/refdata-cellranger-vdj-GRCh38-alts-ensembl-7.0.0.tar.gz
    wget https://cf.10xgenomics.com/supp/cell-vdj/refdata-cellranger-vdj-GRCm38-alts-ensembl-7.0.0.tar.gz

    tar -xzvf cellranger-7.0.1.tar.gz
    tar -xzvf refdata-cellranger-vdj-GRCh38-alts-ensembl-7.0.0.tar.gz
    tar -xzvf refdata-cellranger-vdj-GRCm38-alts-ensembl-7.0.0.tar.gz
    ```

    ## Usage
    
    ```
    conda activate celescope
    multi_flv_CR \\
        --mapfile ./test.mapfile \\
        --thread 8 \\
        --seqtype TCR \\
        --ref_path "/soft/cellranger/vdj_ref/6.0.0/hs/refdata-cellranger-vdj-GRCh38-alts-ensembl-7.0.0" \\
        --soft_path "/soft/cellranger/cellranger-7.0.1/cellranger" \\
        --mod shell
    ```
    """

    def convert(self, sample):
        step = 'convert'
        cmd_line = self.get_cmd_line(step, sample)
        fq2 = f'{self.outdir_dic[sample]["barcode"]}/{sample}_2.fq'
        cmd = (
            f'{cmd_line} '
            f'--fq2 {fq2} '
        )
        self.process_cmd(cmd, step, sample, m=5, x=1)

    def assemble(self, sample):
        step = 'assemble'
        cmd_line = self.get_cmd_line(step, sample)
        fqs_dir = f'{self.outdir_dic[sample]["convert"]}'
        cmd = (
            f'{cmd_line} '
            f'--fqs_dir {fqs_dir} '
        )
        self.process_cmd(cmd, step, sample, m=self.args.mem, x=self.args.thread)

    def summarize(self, sample):
        step = 'summarize'
        cmd_line = self.get_cmd_line(step, sample)
        barcode_convert_json = f'{self.outdir_dic[sample]["convert"]}/barcode_convert.json'
        assemble_out = f'{self.outdir_dic[sample]["assemble"]}/{sample}/outs'
        cmd = (
            f'{cmd_line} '
            f'--barcode_convert_json {barcode_convert_json} '
            f'--assemble_out {assemble_out} '
        )
        self.process_cmd(cmd, step, sample, m=8, x=self.args.thread)

    def match(self, sample):
        step = 'match'
        cmd_line = self.get_cmd_line(step, sample)
        summarize_out = f'{self.outdir_dic[sample]["summarize"]}'
        match_dir = f'{self.col4_dict[sample]}'
        cmd = (
            f'{cmd_line} '
            f'--match_dir {match_dir} '
            f'--summarize_out {summarize_out} '
        )
        self.process_cmd(cmd, step, sample, m=8, x= self.args.thread)
    
    def mapping(self,sample):
        step = 'mapping'
        cmd_line = self.get_cmd_line(step, sample)
        match_dir = f'{self.col4_dict[sample]}'
        match_out = f'{self.outdir_dic[sample]["match"]}'
        cmd = (
            f'{cmd_line} '
            f'--match_dir {match_dir} '
            f'--match_out {match_out} '
        )
        self.process_cmd(cmd, step, sample, m=5, x=1)
    
    def refine_vdj(self, sample):
        step = 'refine_vdj'
        cmd_line = self.get_cmd_line(step, sample)
        match_dir = f'{self.col4_dict[sample]}'
        barcode_convert_json = f'{self.outdir_dic[sample]["convert"]}/barcode_convert.json'
        assemble_out = f'{self.outdir_dic[sample]["assemble"]}/{sample}/outs'
        cmd = (
            f'{cmd_line} '
            f'--match_dir {match_dir} '
            f'--assemble_out {assemble_out} '
            f'--barcode_convert_json {barcode_convert_json} '
        )
        self.process_cmd(cmd, step, sample, m=5, x=1)
            
    def merge_report(self):
        step = "merge_report"
        _index = self.STEPS.index('assemble') + 1
        steps_str = ",".join(self.STEPS[:_index] + __SUB_STEPS__ + self.STEPS[_index:-1])
        samples = ','.join(self.fq_dict.keys())
        app = TOOLS_DIR + '/merge_table.py'
        cmd = (
            f'python {app} --samples {samples} '
            f'--steps {steps_str} --outdir {self.args.outdir}'
        )
        if self.args.rm_files:
            cmd += ' --rm_files'
        self.generate_cmd(cmd, step, sample="")
        for sample in self.fq_dict:
            self.sjm_order += f'order {step} after {self.last_step}_{sample}\n'

def main():
    multi = Multi_flv_CR(__ASSAY__)
    multi.run()
    

if __name__ == '__main__':
    main()




from xopen import xopen
from celescope.tools import utils

import os
import subprocess
import pysam
import pandas as pd


class Filter_noise:
    """ Filter noise barcode.
    For is_cell=False barcodes, keep the highest UMI chain when its UMI >= coeff * second largest UMI. Then only paired-chain barcodes are keeped.

    First sort in descending order according to barcode, chain, umis, so that the next step does not need to be sorted Perform groupby by barcode and chain.
    If each groupby object has only one data, keep it; if it is more than 1, check whether the highest umi is greater than the second highest umi*1.5, if it is satisfied, keep the highest one, if not, then all not reserved Perform groupby by barcode. 
    If the chain number of each barcode is not equal to 2, it will not be retained.
    When retaining, use a set to record the contig_id, and after all data checks are completed, determine which rows need to be retained according to the contig_id.
    """
    
    def __init__(self, df, coeff, seqtype):

        self.df = df.sort_values("umis", ascending=False)
        self.coeff = coeff
        self.seqtype = seqtype
    
    @utils.add_log
    def __call__(self):
        
        df_sorted = self.df.sort_values(['barcode', 'chain', 'umis'], ascending=[False, False, False])

        # 按barcode和chain进行groupby，保留满足条件的数据
        grouped = df_sorted.groupby(['barcode', 'chain'])
        filtered_data = pd.DataFrame(columns=df_sorted.columns)  # 创建一个空的DataFrame来存储过滤后的数据
        contig_id_set = set()  # 用于记录需要保留的contig_id

        for _group_name, group_data in grouped:
            if len(group_data) == 1:
                contig_id_set.add(group_data['contig_id'].values[0])  
            else:
                sorted_umis = group_data['umis'].values
                highest_umi = sorted_umis[0]
                second_highest_umi = sorted_umis[1]
                if highest_umi >= second_highest_umi * self.coeff:
                    contig_id_set.add(group_data.iloc[0]['contig_id'])  # 记录保留的contig_id

        filtered_data = self.df[self.df['contig_id'].isin(contig_id_set)]

        grouped_barcode = filtered_data.groupby('barcode')
        final_filtered_data = pd.DataFrame(columns=filtered_data.columns)
        contig_id_set = set()

        for _group_name, group_data in grouped_barcode:
            if len(group_data['chain'].unique()) == 2:
                if self.seqtype == 'BCR' and 'IGH' not in group_data['chain'].unique():
                    continue
                contig_id_set.update(set(group_data['contig_id']))  # 记录保留的contig_id

        # 根据contig_id确定需要保留的行
        final_filtered_data = filtered_data[filtered_data['contig_id'].isin(contig_id_set)]
        final_filtered_data = final_filtered_data.sort_values(['barcode'])

        return final_filtered_data


class Refine_assemble:
    """
    ## Features

    - TCR/BCR Assemble by Cellranger.

    - Generate Mapping, Cells, V(D)J annotations metrics in html.

    ## Output
    
    - `03.assemble/{sample}_refine` Cellranger vdj refined results.

    """
    def __init__(self, args):

        # common
        self.sample = args.sample
        self.outdir = args.outdir
        self.coeff = args.coeff
        self.seqtype = args.seqtype
        
        # in
        self.raw_all_contig_anno = pd.read_csv(f"{self.outdir}/{self.sample}/outs/all_contig_annotations.csv")
        self.raw_filter_contig_anno = pd.read_csv(f"{self.outdir}/{self.sample}/outs/filtered_contig_annotations.csv")
        self.all_bam = f"{self.outdir}/{self.sample}/outs/all_contig.bam"
        self.fqs_dir = os.path.abspath(args.fqs_dir)
        self.raw_convert_fq1 = f'{self.fqs_dir}/{self.sample}_S1_L001_R1_001.fastq.gz'
        self.raw_convert_fq2 = f'{self.fqs_dir}/{self.sample}_S1_L001_R2_001.fastq.gz'
        
        # rename sample
        self.refine_sample = self.sample + "_refine"
        
        # out
        self.refine_fqs_dir = os.path.abspath(f"{self.outdir}/refine_convert")
        utils.check_mkdir(self.refine_fqs_dir)
        
        self.out_fq1_file = f'{self.refine_fqs_dir}/{self.refine_sample}_S1_L001_R1_001.fastq.gz'
        self.out_fq2_file = f'{self.refine_fqs_dir}/{self.refine_sample}_S1_L001_R2_001.fastq.gz'
        self.cmd_line = f'{self.outdir}/{self.sample}_refine_cmd_line'
        
        # running parameters
        self.mem = args.mem
        self.thread = args.thread
        self.other_param = args.other_param
        self.ref_path = args.ref_path
        self.soft_path = args.soft_path

    
    @utils.add_log
    def refine_vdj(self):
        """return two sets：
        1. successfully assembled barcode sets.
        2. refined contig_id which including paired productive contig and all non-productive contig_id.
        """
        df_productive = self.raw_all_contig_anno[self.raw_all_contig_anno["productive"] == True]
        df_refine = df_productive[df_productive["is_cell"] == False]
        df_non_productive = df_refine[df_refine["productive"] == False]
        
        df_refine = df_refine.groupby("barcode").filter(lambda x: (len(x) > 4))
        df_refine = Filter_noise(df_refine, self.coeff, self.seqtype)()
        refined_contigs = set(df_refine.contig_id)
        refined_contigs |= set(df_non_productive.contig_id)

        df_productive = self.raw_filter_contig_anno[self.raw_filter_contig_anno["productive"] == True]
        raw_assembled_cells = set(df_productive.barcode)
        
        return raw_assembled_cells, refined_contigs


    @utils.add_log
    def gen_refine_fqs(self, raw_assembled_cells, refined_contigs):
        # some reads may not map to assembled contigs but used for assembling
        refined_contigs.add(None)
        read_name_set = set()

        with pysam.AlignmentFile(self.all_bam) as f:
            for read in f:
                # 记录所有被判定为细胞的contig的reads，确保不影响已经成功组装的细胞数
                if read.get_tag("CB") in raw_assembled_cells:
                    read_name_set.add(read.query_name)
                # 记录refined contig的reads
                elif read.reference_name in refined_contigs:
                    read_name_set.add(read.query_name)

        out_fq1 = xopen(self.out_fq1_file, 'w')
        out_fq2 = xopen(self.out_fq2_file, 'w')

        with pysam.FastxFile(self.raw_convert_fq1, persist=False) as fq1, \
                pysam.FastxFile(self.raw_convert_fq2, persist=False) as fq2:
            for entry1, entry2 in zip(fq1, fq2):
                header1, seq1, qual1 = entry1.name, entry1.sequence, entry1.quality
                header2, seq2, qual2 = entry2.name, entry2.sequence, entry2.quality
                if header1 in read_name_set:
                    out_fq1.write(f"@{header1}\n{seq1}\n+\n{qual1}\n")
                    out_fq2.write(f'@{header2}\n{seq2}\n+\n{qual2}\n')

        out_fq1.close()
        out_fq2.close() 

    
    @utils.add_log
    def refine_assemble(self):
        """Cellranger vdj"""
        cmd = (
            f'{self.soft_path} vdj '
            f'--id={self.refine_sample} '
            f'--reference={self.ref_path} '
            f'--fastqs={self.refine_fqs_dir} '
            f'--sample={self.refine_sample} '
            f'--localcores={self.thread} '
            f'--localmem={self.mem} '
        )

        if self.other_param:
            cmd += (" " + self.other_param)

        with open(self.cmd_line, 'w') as f:
            f.write(cmd)
        
        cwd = os.getcwd()
        os.chdir(self.outdir)
        subprocess.check_call(cmd, shell=True)
        # change dir back to avoid can not find '03.assemble/stat.txt' error
        os.chdir(cwd)

    def run(self):
        raw_assembled_cells, refined_contigs = self.refine_vdj()
        self.gen_refine_fqs(raw_assembled_cells, refined_contigs)
        self.refine_assemble()